@@ -13,7 +13,7 @@
 //! Behavior of `derive(Options)` can be controlled by adding `#[options(...)]`
 //! attributes to one or more fields within a decorated struct.
 //!
-//! Supported items are:
+//! Supported items for `struct` fields are:
 //!
 //! * `command` indicates that a field represents a subcommand. The field must
 //!   be of type `Option<T>` where `T` is a type implementing `Options`.
@@ -35,7 +35,8 @@
 //!   The value of this field is parsed in the same way as argument values.
 //! * `default_expr` provides a default value for the option field.
 //!   The value of this field is parsed at compile time as a Rust expression
-//!   and is evaluated before any argument values are processed.
+//!   and is evaluated before any argument values are processed.  
+//!   The `default_expr` feature must be enabled to use this attribute.
 //! * `required` will cause an error if the option is not present,
 //!   unless at least one `help_flag` option is also present.
 //! * `multi = "..."` will allow parsing an option multiple times,
@@ -57,35 +58,48 @@
 //!     * `parse(from_str)` uses `std::convert::From::from`
 //!     * `parse(try_from_str)` uses `std::str::FromStr::from_str`
 //!
-//! The `options` attribute may also be added at the type level.
+//! Additionally, the following flags may be set at the type level to establish
+//! default values for all contained fields: `no_help_flag`, `no_long`,
+//! `no_short`, and `required`.
+//!
+//! Supported items for `enum` variants are:
+//!
+//! * `name = "..."` sets the user-facing command name.  
+//!   If this option is not present, one is automatically generated from the variant name.
+//! * `help = "..."` sets the help string for the command;
+//!   variant doc comment may also be provided to set the help text.
+//!   If both are present, the `help` attribute value is used.
 //!
 //! The `help` attribute (or a type-level doc comment) can be used to provide
 //! some introductory text which will precede option help text in the usage
 //! string.
-//!
-//! Additionally, the following flags may be set at the type level to establish
-//! default values for all contained fields: `no_help_flag`, `no_long`,
-//! `no_short`, and `required`.
 
 #![recursion_limit = "1024"]
 
 extern crate proc_macro;
-extern crate proc_macro2;
-#[macro_use] extern crate quote;
-extern crate syn;
 
 use std::iter::repeat;
 
+use quote::quote;
+
 use proc_macro::TokenStream;
 use proc_macro2::{Span, TokenStream as TokenStream2};
 
 use syn::{
-    parse::Error, spanned::Spanned, Expr,
+    parse::Error, spanned::Spanned,
     Attribute, AttrStyle, Data, DataEnum, DataStruct, DeriveInput, Fields,
     GenericArgument, Ident, Lit, Meta, NestedMeta, Path, PathArguments, Type,
     parse_str,
 };
 
+#[cfg(feature = "default_expr")]
+use syn::Expr;
+
+/// Derives the `gumdrop::Options` trait for `struct` and `enum` items.
+///
+/// `#[options(...)]` attributes can be used to control behavior of generated trait
+/// implementation. See [crate-level documentation](index.html) for the full list of
+/// supported options.
 #[proc_macro_derive(Options, attributes(options))]
 pub fn derive_options(input: TokenStream) -> TokenStream {
     let ast: DeriveInput = match syn::parse(input) {
@@ -133,7 +147,7 @@ fn derive_options_enum(ast: &DeriveInput, data: &DataEnum)
                 return Err(Error::new(span,
                     "command variants must be unary tuple variants")),
             Fields::Unnamed(fields) =>
-                &fields.unnamed.first().unwrap().into_value().ty,
+                &fields.unnamed.first().unwrap().ty,
         };
 
         let opts = CmdOpts::parse(&var.attrs)?;
@@ -179,6 +193,16 @@ fn derive_options_enum(ast: &DeriveInput, data: &DataEnum)
 
     let (impl_generics, ty_generics, where_clause) = ast.generics.split_for_impl();
 
+    let command_impl = {
+        let name = repeat(name);
+
+        quote!{
+            match self {
+                #( #name::#variant(cmd) => ::gumdrop::Options::command(cmd), )*
+            }
+        }
+    };
+
     let command_name_impl = {
         let name = repeat(name);
 
@@ -189,6 +213,26 @@ fn derive_options_enum(ast: &DeriveInput, data: &DataEnum)
         }
     };
 
+    let self_usage_impl = {
+        let name = repeat(name);
+
+        quote!{
+            match self {
+                #( #name::#variant(sub) => ::gumdrop::Options::self_usage(sub), )*
+            }
+        }
+    };
+
+    let self_command_list_impl = {
+        let name = repeat(name);
+
+        quote!{
+            match self {
+                #( #name::#variant(sub) => ::gumdrop::Options::self_command_list(sub), )*
+            }
+        }
+    };
+
     Ok(quote!{
         impl #impl_generics ::gumdrop::Options for #name #ty_generics #where_clause {
             fn parse<__S: ::std::convert::AsRef<str>>(
@@ -200,6 +244,10 @@ fn derive_options_enum(ast: &DeriveInput, data: &DataEnum)
                 Self::parse_command(_arg, _parser)
             }
 
+            fn command(&self) -> ::std::option::Option<&dyn ::gumdrop::Options> {
+                #command_impl
+            }
+
             fn command_name(&self) -> ::std::option::Option<&'static str> {
                 #command_name_impl
             }
@@ -226,10 +274,18 @@ fn derive_options_enum(ast: &DeriveInput, data: &DataEnum)
                 #usage
             }
 
+            fn self_usage(&self) -> &'static str {
+                #self_usage_impl
+            }
+
             fn command_list() -> ::std::option::Option<&'static str> {
                 ::std::option::Option::Some(<Self as ::gumdrop::Options>::usage())
             }
 
+            fn self_command_list(&self) -> ::std::option::Option<&'static str> {
+                #self_command_list_impl
+            }
+
             fn command_usage(name: &str) -> ::std::option::Option<&'static str> {
                 match name {
                     #( #command => ::std::option::Option::Some(
@@ -275,10 +331,18 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
             default.push(opts.parse.as_ref()
                 .unwrap_or(&ParseFn::Default)
                 .make_parse_default_action(ident, &expr));
-        } else if let Some(expr) = &opts.default_expr {
-            default.push(quote!{ #expr });
         } else {
+            #[cfg(not(feature = "default_expr"))]
             default.push(default_expr.clone());
+
+            #[cfg(feature = "default_expr")]
+            {
+                if let Some(expr) = &opts.default_expr {
+                    default.push(quote!{ #expr });
+                } else {
+                    default.push(default_expr.clone());
+                }
+            }
         }
 
         if opts.command {
@@ -529,6 +593,15 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
         }
     };
 
+    let command_impl = match &command {
+        None => quote!{ ::std::option::Option::None },
+        Some(field) => quote!{
+            ::std::option::Option::map(
+                ::std::option::Option::as_ref(&self.#field),
+                |sub| sub as _)
+        }
+    };
+
     let command_name_impl = match &command {
         None => quote!{ ::std::option::Option::None },
         Some(field) => quote!{
@@ -573,6 +646,26 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
         }
     };
 
+    let self_usage_impl = match &command {
+        None => quote!{ <Self as ::gumdrop::Options>::usage() },
+        Some(field) => quote!{
+            ::std::option::Option::map_or_else(
+                ::std::option::Option::as_ref(&self.#field),
+                <Self as ::gumdrop::Options>::usage,
+                ::gumdrop::Options::self_usage)
+        }
+    };
+
+    let self_command_list_impl = match &command {
+        None => quote!{ <Self as ::gumdrop::Options>::command_list() },
+        Some(field) => quote!{
+            ::std::option::Option::map_or_else(
+                ::std::option::Option::as_ref(&self.#field),
+                <Self as ::gumdrop::Options>::command_list,
+                ::gumdrop::Options::self_command_list)
+        }
+    };
+
     let required = &required;
 
     let (impl_generics, ty_generics, where_clause) = ast.generics.split_for_impl();
@@ -615,6 +708,10 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
                 ::std::result::Result::Ok(_result)
             }
 
+            fn command(&self) -> ::std::option::Option<&dyn ::gumdrop::Options> {
+                #command_impl
+            }
+
             fn command_name(&self) -> ::std::option::Option<&'static str> {
                 #command_name_impl
             }
@@ -632,6 +729,10 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
                 #usage
             }
 
+            fn self_usage(&self) -> &'static str {
+                #self_usage_impl
+            }
+
             fn command_list() -> ::std::option::Option<&'static str> {
                 #command_list
             }
@@ -639,6 +740,10 @@ fn derive_options_struct(ast: &DeriveInput, fields: &Fields)
             fn command_usage(_name: &str) -> ::std::option::Option<&'static str> {
                 #command_usage
             }
+
+            fn self_command_list(&self) -> ::std::option::Option<&'static str> {
+                #self_command_list_impl
+            }
         }
     })
 }
@@ -675,6 +780,7 @@ struct AttrOpts {
     meta: Option<String>,
     parse: Option<ParseFn>,
     default: Option<String>,
+    #[cfg(feature = "default_expr")]
     default_expr: Option<Expr>,
 
     command: bool,
@@ -749,7 +855,7 @@ impl Action {
     fn infer(ty: &Type, opts: &AttrOpts) -> Action {
         match ty {
             Type::Path(path) => {
-                let path = path.path.segments.last().unwrap().into_value();
+                let path = path.path.segments.last().unwrap();
                 let param = first_ty_param(ty);
 
                 match &path.ident.to_string()[..] {
@@ -885,8 +991,11 @@ impl AttrOpts {
             if self.count { err!("`count` and `parse` are mutually exclusive"); }
         }
 
-        if self.default.is_some() && self.default_expr.is_some() {
-            err!("`default` and `default_expr` are mutually exclusive");
+        #[cfg(feature = "default_expr")]
+        {
+            if self.default.is_some() && self.default_expr.is_some() {
+                err!("`default` and `default_expr` are mutually exclusive");
+            }
         }
 
         Ok(())
@@ -911,11 +1020,11 @@ impl AttrOpts {
                     let meta = attr.parse_meta()?;
 
                     match meta {
-                        Meta::Word(ident) =>
-                            return Err(Error::new(ident.span(),
+                        Meta::Path(path) =>
+                            return Err(Error::new(path.span(),
                                 "`#[options]` is not a valid attribute")),
                         Meta::NameValue(nv) =>
-                            return Err(Error::new(nv.ident.span(),
+                            return Err(Error::new(nv.path.span(),
                                 "`#[options = ...]` is not a valid attribute")),
                         Meta::List(items) => {
                             for item in &items.nested {
@@ -934,51 +1043,64 @@ impl AttrOpts {
 
     fn parse_item(&mut self, item: &NestedMeta) -> Result<(), Error> {
         match item {
-            NestedMeta::Literal(lit) =>
+            NestedMeta::Lit(lit) =>
                 return Err(unexpected_meta_item(lit.span())),
             NestedMeta::Meta(item) => {
                 match item {
-                    Meta::Word(w) => match &w.to_string()[..] {
-                        "free" => self.free = true,
-                        "command" => self.command = true,
-                        "count" => self.count = true,
-                        "help_flag" => self.help_flag = true,
-                        "no_help_flag" => self.no_help_flag = true,
-                        "no_short" => self.no_short = true,
-                        "no_long" => self.no_long = true,
-                        "no_multi" => self.no_multi = true,
-                        "required" => self.required = true,
-                        "not_required" => self.not_required = true,
-                        _ => return Err(unexpected_meta_item(w.span()))
+                    Meta::Path(path) => match path.get_ident() {
+                        Some(ident) => match ident.to_string().as_str() {
+                            "free" => self.free = true,
+                            "command" => self.command = true,
+                            "count" => self.count = true,
+                            "help_flag" => self.help_flag = true,
+                            "no_help_flag" => self.no_help_flag = true,
+                            "no_short" => self.no_short = true,
+                            "no_long" => self.no_long = true,
+                            "no_multi" => self.no_multi = true,
+                            "required" => self.required = true,
+                            "not_required" => self.not_required = true,
+                            _ => return Err(unexpected_meta_item(path.span()))
+                        }
+                        None => return Err(unexpected_meta_item(path.span()))
                     },
                     Meta::List(list) => {
-                        match &list.ident.to_string()[..] {
-                            "parse" => {
+                        match list.path.get_ident() {
+                            Some(ident) if ident.to_string() == "parse" => {
                                 if list.nested.len() != 1 {
-                                    return Err(unexpected_meta_item(list.ident.span()));
+                                    return Err(unexpected_meta_item(list.path.span()));
                                 }
 
                                 self.parse = Some(ParseFn::parse(&list.nested[0])?);
                             }
-                            _ => return Err(unexpected_meta_item(list.ident.span()))
+                            _ => return Err(unexpected_meta_item(list.path.span()))
                         }
                     }
                     Meta::NameValue(nv) => {
-                        match &nv.ident.to_string()[..] {
-                            "default" => self.default = Some(lit_str(&nv.lit)?),
-                            "default_expr" => {
-                                let expr = parse_str(&lit_str(&nv.lit)?)?;
-                                self.default_expr = Some(expr);
+                        match nv.path.get_ident() {
+                            Some(ident) => match ident.to_string().as_str() {
+                                "default" => self.default = Some(lit_str(&nv.lit)?),
+                                #[cfg(feature = "default_expr")]
+                                "default_expr" => {
+                                    let expr = parse_str(&lit_str(&nv.lit)?)?;
+                                    self.default_expr = Some(expr);
+                                }
+                                #[cfg(not(feature = "default_expr"))]
+                                "default_expr" => {
+                                    return Err(Error::new(nv.path.span(),
+                                    "compile gumdrop with the `default_expr` \
+                                    feature to enable this attribute"));
+                                }
+                                "long" => self.long = Some(lit_str(&nv.lit)?),
+                                "short" => self.short = Some(lit_char(&nv.lit)?),
+                                "help" => self.help = Some(lit_str(&nv.lit)?),
+                                "meta" => self.meta = Some(lit_str(&nv.lit)?),
+                                "multi" => {
+                                    let name = parse_str(&lit_str(&nv.lit)?)?;
+                                    self.multi = Some(name);
+                                }
+                                _ => return Err(unexpected_meta_item(nv.path.span()))
                             }
-                            "long" => self.long = Some(lit_str(&nv.lit)?),
-                            "short" => self.short = Some(lit_char(&nv.lit)?),
-                            "help" => self.help = Some(lit_str(&nv.lit)?),
-                            "meta" => self.meta = Some(lit_str(&nv.lit)?),
-                            "multi" => {
-                                let name = parse_str(&lit_str(&nv.lit)?)?;
-                                self.multi = Some(name);
-                            }
-                            _ => return Err(unexpected_meta_item(nv.ident.span()))
+                            None => return Err(unexpected_meta_item(nv.path.span()))
                         }
                     }
                 }
@@ -1030,11 +1152,11 @@ impl CmdOpts {
                     let meta = attr.parse_meta()?;
 
                     match meta {
-                        Meta::Word(ident) =>
-                            return Err(Error::new(ident.span(),
+                        Meta::Path(path) =>
+                            return Err(Error::new(path.span(),
                                 "`#[options]` is not a valid attribute")),
                         Meta::NameValue(nv) =>
-                            return Err(Error::new(nv.ident.span(),
+                            return Err(Error::new(nv.path.span(),
                                 "`#[options = ...]` is not a valid attribute")),
                         Meta::List(items) => {
                             for item in &items.nested {
@@ -1051,19 +1173,22 @@ impl CmdOpts {
 
     fn parse_item(&mut self, item: &NestedMeta) -> Result<(), Error> {
         match item {
-            NestedMeta::Literal(lit) =>
+            NestedMeta::Lit(lit) =>
                 return Err(unexpected_meta_item(lit.span())),
             NestedMeta::Meta(item) => {
                 match item {
-                    Meta::Word(ident) =>
-                        return Err(unexpected_meta_item(ident.span())),
+                    Meta::Path(path) =>
+                        return Err(unexpected_meta_item(path.span())),
                     Meta::List(list) =>
-                        return Err(unexpected_meta_item(list.ident.span())),
+                        return Err(unexpected_meta_item(list.path.span())),
                     Meta::NameValue(nv) => {
-                        match &nv.ident.to_string()[..] {
-                            "name" => self.name = Some(lit_str(&nv.lit)?),
-                            "help" => self.help = Some(lit_str(&nv.lit)?),
-                            _ => return Err(unexpected_meta_item(nv.ident.span()))
+                        match nv.path.get_ident() {
+                            Some(ident) => match ident.to_string().as_str() {
+                                "name" => self.name = Some(lit_str(&nv.lit)?),
+                                "help" => self.help = Some(lit_str(&nv.lit)?),
+                                _ => return Err(unexpected_meta_item(nv.path.span()))
+                            }
+                            None => return Err(unexpected_meta_item(nv.path.span()))
                         }
                     }
                 }
@@ -1086,7 +1211,10 @@ impl DefaultOpts {
                     if let Meta::NameValue(nv) = meta {
                         let doc = lit_str(&nv.lit)?;
 
-                        if opts.doc.is_none() {
+                        if let Some(text) = opts.doc.as_mut() {
+                            text.push('\n');
+                            text.push_str(doc.trim_start());
+                        } else {
                             opts.doc = Some(doc.trim_start().to_owned());
                         }
                     }
@@ -1094,11 +1222,11 @@ impl DefaultOpts {
                     let meta = attr.parse_meta()?;
 
                     match meta {
-                        Meta::Word(ident) =>
-                            return Err(Error::new(ident.span(),
+                        Meta::Path(path) =>
+                            return Err(Error::new(path.span(),
                                 "`#[options]` is not a valid attribute")),
                         Meta::NameValue(nv) =>
-                            return Err(Error::new(nv.ident.span(),
+                            return Err(Error::new(nv.path.span(),
                                 "`#[options = ...]` is not a valid attribute")),
                         Meta::List(items) => {
                             for item in &items.nested {
@@ -1115,26 +1243,29 @@ impl DefaultOpts {
 
     fn parse_item(&mut self, item: &NestedMeta) -> Result<(), Error> {
         match item {
-            NestedMeta::Literal(lit) =>
+            NestedMeta::Lit(lit) =>
                 return Err(unexpected_meta_item(lit.span())),
             NestedMeta::Meta(item) => {
                 match item {
-                    Meta::Word(w) => match &w.to_string()[..] {
-                        "no_help_flag" => self.no_help_flag = true,
-                        "no_short" => self.no_short = true,
-                        "no_long" => self.no_long = true,
-                        "no_multi" => self.no_multi = true,
-                        "required" => self.required = true,
-                        _ => return Err(unexpected_meta_item(w.span()))
+                    Meta::Path(path) => match path.get_ident() {
+                        Some(ident) => match ident.to_string().as_str() {
+                            "no_help_flag" => self.no_help_flag = true,
+                            "no_short" => self.no_short = true,
+                            "no_long" => self.no_long = true,
+                            "no_multi" => self.no_multi = true,
+                            "required" => self.required = true,
+                            _ => return Err(unexpected_meta_item(ident.span()))
+                        }
+                        None => return Err(unexpected_meta_item(path.span()))
                     },
                     Meta::NameValue(nv) => {
-                        match &nv.ident.to_string()[..] {
-                            "help" => self.help = Some(lit_str(&nv.lit)?),
-                            _ => return Err(unexpected_meta_item(nv.ident.span()))
+                        match nv.path.get_ident() {
+                           Some(ident) if ident.to_string() == "help" => self.help = Some(lit_str(&nv.lit)?),
+                            _ => return Err(unexpected_meta_item(nv.path.span()))
                         }
                     }
                     Meta::List(list) =>
-                        return Err(unexpected_meta_item(list.ident.span()))
+                        return Err(unexpected_meta_item(list.path.span()))
                 }
             }
         }
@@ -1147,7 +1278,7 @@ impl FreeAction {
     fn infer(ty: &Type, opts: &AttrOpts) -> FreeAction {
         match ty {
             Type::Path(path) => {
-                let path = path.path.segments.last().unwrap().into_value();
+                let path = path.path.segments.last().unwrap();
 
                 match &path.ident.to_string()[..] {
                     "Option" => FreeAction::SetOption,
@@ -1344,27 +1475,33 @@ impl<'a> Opt<'a> {
 impl ParseFn {
     fn parse(item: &NestedMeta) -> Result<ParseFn, Error> {
         let result = match item {
-            NestedMeta::Meta(Meta::Word(ident)) => {
-                match &ident.to_string()[..] {
-                    "from_str" => ParseFn::FromStr(None),
-                    "try_from_str" => ParseFn::Default,
-                    _ => return Err(unexpected_meta_item(ident.span()))
+            NestedMeta::Meta(Meta::Path(path)) => {
+                match path.get_ident() {
+                    Some(ident) => match ident.to_string().as_str() {
+                        "from_str" => ParseFn::FromStr(None),
+                        "try_from_str" => ParseFn::Default,
+                        _ => return Err(unexpected_meta_item(ident.span()))
+                    }
+                    None => return Err(unexpected_meta_item(path.span()))
                 }
             }
             NestedMeta::Meta(Meta::NameValue(nv)) => {
-                match &nv.ident.to_string()[..] {
-                    "from_str" => {
-                        let path = parse_str(&lit_str(&nv.lit)?)?;
-                        ParseFn::FromStr(Some(path))
+                match nv.path.get_ident() {
+                    Some(ident) => match ident.to_string().as_str() {
+                        "from_str" => {
+                            let path = parse_str(&lit_str(&nv.lit)?)?;
+                            ParseFn::FromStr(Some(path))
+                        }
+                        "try_from_str" => {
+                            let path = parse_str(&lit_str(&nv.lit)?)?;
+                            ParseFn::TryFromStr(path)
+                        }
+                        _ => return Err(unexpected_meta_item(nv.path.span()))
                     }
-                    "try_from_str" => {
-                        let path = parse_str(&lit_str(&nv.lit)?)?;
-                        ParseFn::TryFromStr(path)
-                    }
-                    _ => return Err(unexpected_meta_item(nv.ident.span()))
+                    None => return Err(unexpected_meta_item(nv.path.span()))
                 }
             }
-            NestedMeta::Literal(_) |
+            NestedMeta::Lit(_) |
             NestedMeta::Meta(Meta::List(_)) =>
                 return Err(unexpected_meta_item(item.span())),
         };
@@ -1480,7 +1617,7 @@ impl ParseMethod {
 fn first_ty_param(ty: &Type) -> Option<&Type> {
     match ty {
         Type::Path(path) => {
-            let path = path.path.segments.last().unwrap().into_value();
+            let path = path.path.segments.last().unwrap();
 
             match &path.arguments {
                 PathArguments::AngleBracketed(data) =>
@@ -1532,7 +1669,7 @@ fn lit_char(lit: &Lit) -> Result<char, Error> {
 
 fn path_eq(path: &Path, s: &str) -> bool {
     path.segments.len() == 1 && {
-        let seg = path.segments.first().unwrap().into_value();
+        let seg = path.segments.first().unwrap();
 
         match seg.arguments {
             PathArguments::None => seg.ident == s,
@@ -1640,13 +1777,17 @@ fn make_usage(help: &Option<String>, free: &[FreeOpt], opts: &[Opt]) -> String {
 
     if let Some(help) = help {
         res.push_str(help);
-        res.push_str("\n\n");
+        res.push('\n');
     }
 
     let width = max_width(free, |opt| opt.width())
         .max(max_width(opts, |opt| opt.width()));
 
     if !free.is_empty() {
+        if !res.is_empty() {
+            res.push('\n');
+        }
+
         res.push_str("Positional arguments:\n");
 
         for opt in free {
